# Imports do próprio módulo
from inewave.config import MAX_ITERS, REES
from inewave._utils.bloco import Bloco
from inewave._utils.leitura import Leitura
# Imports de módulos externos
import numpy as np  # type: ignore
from typing import IO, List


class RegistroEstado:
    """
    Armazena as informações de um registro visitado pelo NEWAVE
    em algum período de alguma simulação.

    ** Parâmetros **

    - ireg: `int`
    - itec: `int`
    - simc: `int`
    - itef: `int`
    - fobj: `int`
    - tabela: `np.ndarray`

    """

    __slots__ = ["ireg",
                 "itec",
                 "simc",
                 "itef",
                 "fobj",
                 "tabela"]

    def __init__(self,
                 ireg: int,
                 itec: int,
                 simc: int,
                 itef: int,
                 fobj: float,
                 tabela: np.ndarray):
        self.ireg = ireg
        self.itec = itec
        self.simc = simc
        self.itef = itef
        self.fobj = fobj
        self.tabela = tabela

    def __eq__(self, o: object) -> bool:
        """
        A igualdade entre RegistroEstado avalia todos os
        valores.
        """
        if not isinstance(o, RegistroEstado):
            return False
        reg: RegistroEstado = o
        eq_ireg = self.ireg == reg.ireg
        eq_itec = self.itec == reg.itec
        eq_simc = self.simc == reg.simc
        eq_itef = self.itef == reg.itef
        eq_fobj = self.fobj == reg.fobj
        eq_tab = np.array_equal(self.tabela, reg.tabela)

        return all([eq_ireg,
                    eq_itec,
                    eq_simc,
                    eq_itef,
                    eq_fobj,
                    eq_tab])

    @classmethod
    def le_registro(cls, primeira_linha: str, arq: IO) -> 'RegistroEstado':
        """
        """
        primeira = True
        n_rees = len(REES)
        n_cols_tabela = 18
        ireg = 0
        itec = 0
        simc = 0
        itef = 0
        fobj = 0.
        tabela = np.zeros((n_rees, n_cols_tabela))
        for i in range(n_rees):
            if primeira:
                primeira = False
                linha = primeira_linha
                # Extrai os campos específicos da primeira linha
                ireg = int(linha[2:10])
                itec = int(linha[11:15])
                simc = int(linha[16:20])
                itef = int(linha[21:25])
                fobj = float(linha[31:48])
            else:
                linha = arq.readline()
            # Preenche a tabela com os dados do registro
            ree = int(linha[26:30])
            ci = 49
            nc = 17
            for j in range(n_cols_tabela):
                cf = ci + nc
                num_str = linha[ci:cf]
                valor = 0.0 if not num_str.isnumeric() else float(num_str)
                tabela[ree-1, j] = valor
                ci = cf + 1

        return cls(ireg,
                   itec,
                   simc,
                   itef,
                   fobj,
                   tabela)


class BlocoRegistroEstados(Bloco):
    """
    Bloco com informações do estado de construção dos cortes
    de um período, existentes no arquivo `estados.rel` do NWLISTCF.
    """
    str_inicio = "  PERIODO:  "

    def __init__(self):

        super().__init__(BlocoRegistroEstados.str_inicio,
                         "",
                         False)

        self._dados: List[RegistroEstado] = []

    def __eq__(self, o: object):
        if not isinstance(o, BlocoRegistroEstados):
            return False
        bloco: BlocoRegistroEstados = o
        return all([d1 == d2
                    for d1, d2 in zip(self._dados, bloco._dados)])

    # Override
    def le(self, arq: IO):
        # Salta duas linhas para acessar a tabela
        arq.readline()
        arq.readline()
        while True:
            # Verifica se a próxima linha é o início do
            # próximo período ou é vazia
            linha = arq.readline()
            if (BlocoRegistroEstados.str_inicio in linha
                    or len(linha) < 2):
                break
            # Senão, lê mais um registro
            reg = RegistroEstado.le_registro(linha, arq)
            self._dados.append(reg)

        return linha

    # Override
    def escreve(self, arq: IO):
        pass


class LeituraEstados(Leitura):
    """
    Realiza a leitura do arquivo estados.rel,
    existente em um diretório de saídas do NWLISTCF.

    Esta classe contém o conjunto de utilidades para ler
    e interpretar os campos do arquivo estados.rel, construindo um
    objeto `Estados` cujas informações são as mesmas do arquivo.

    Este objeto existe para retirar do modelo de dados a complexidade
    de iterar pelas linhas do arquivo, recortar colunas, converter
    tipos de dados, dentre outras tarefas necessárias para a leitura.

    """
    str_inicio_periodo = "  PERIODO:   "

    def __init__(self,
                 diretorio: str) -> None:
        super().__init__(diretorio)

    def _cria_blocos_leitura(self) -> List[Bloco]:
        return [BlocoRegistroEstados()
                for _ in range(MAX_ITERS)]



# Imports do próprio módulo
from inewave._utils.registros import RegistroFn, RegistroIn
from inewave._utils.bloco import Bloco
from inewave._utils.leitura import Leitura
from inewave.config import MAX_ANOS_ESTUDO
from inewave.config import MESES
from inewave.config import MESES_DF
from inewave.config import NUM_PATAMARES
from inewave.config import SUBMERCADOS

# Imports de módulos externos
import numpy as np  # type: ignore
import pandas as pd  # type: ignore
from typing import IO, List


class BlocoDuracaoPatamar(Bloco):
    """
    Bloco com a duração de cada patamar por mês
    de estudo, extraído do arquivo `patamar.dat`.
    """
    str_inicio = "ANO   DURACAO MENSAL DOS PATAMARES DE CARGA"
    str_fim = "SUBSISTEMA"

    def __init__(self):

        super().__init__(BlocoDuracaoPatamar.str_inicio,
                         "",
                         True)

        self._dados: pd.DataFrame = pd.DataFrame()

    def __eq__(self, o: object):
        if not isinstance(o, BlocoDuracaoPatamar):
            return False
        bloco: BlocoDuracaoPatamar = o
        return self._dados.equals(bloco._dados)

    # Override
    def le(self, arq: IO):

        def converte_tabela_em_df() -> pd.DataFrame:
            pats = [n for n in range(1, NUM_PATAMARES + 1)]
            coluna_pats = pats * int(i / NUM_PATAMARES)
            df = pd.DataFrame(tabela)
            df.columns = MESES_DF
            df["Ano"] = anos
            df["Patamar"] = coluna_pats
            df = df[["Ano", "Patamar"] + MESES_DF]
            return df

        # Variáveis auxiliares
        reg_ano = RegistroIn(4)
        reg_pat = RegistroFn(6)
        # Pula as duas primeiras linhas, com cabeçalhos
        arq.readline()
        arq.readline()
        i = 0
        anos = []
        tabela = np.zeros((MAX_ANOS_ESTUDO * NUM_PATAMARES,
                          len(MESES)))
        while True:
            # Verifica se o arquivo acabou
            linha = arq.readline()
            if BlocoDuracaoPatamar.str_fim in linha:
                tabela = tabela[:i, :]
                self._dados = converte_tabela_em_df()
                break
            # Senão, lê mais uma linha
            # Ano
            if linha[0:4].isnumeric():
                anos.append(reg_ano.le_registro(linha, 0))
            else:
                anos.append(anos[-1])
            # Patamares
            tabela[i, :] = reg_pat.le_linha_tabela(linha,
                                                   6,
                                                   2,
                                                   len(MESES))
            i += 1

    # Override
    def escreve(self, arq: IO):

        def escreve_patamares():
            lin_tab = self._dados.shape[0]
            for i in range(lin_tab):
                linha = ""
                # Ano
                if i % NUM_PATAMARES == 0:
                    linha += str(int(self._dados.iloc[i, 0])).rjust(4)
                else:
                    linha += "    "
                # Patamares de cada mês
                for j in range(len(MESES)):
                    v = self._dados.iloc[i, j + 2]
                    linha += "{:1.4f}".format(v).rjust(8)
                arq.write(linha + "\n")

        # Escreve cabeçalhos
        titulos = ("      JAN     FEV     MAR     ABR     MAI     JUN     "
                   + "JUL     AGO     SET     OUT     NOV     DEZ" + "\n")
        cab = ("      X.XXXX  X.XXXX  X.XXXX  X.XXXX  X.XXXX  X.XXXX  "
               + "X.XXXX  X.XXXX  X.XXXX  X.XXXX  X.XXXX  X.XXXX" + "\n")
        arq.write(f"{BlocoDuracaoPatamar.str_inicio}\n")
        arq.write(titulos)
        arq.write(cab)
        escreve_patamares()
        # Escreve a linha de terminação
        arq.write(f" {BlocoDuracaoPatamar.str_fim}\n")


class BlocoCargaPatamarSubsistemas(Bloco):
    """
    Bloco com a informação de carga (em p.u.) por patamar de carga
    e por mês/ano de estudo para cada subsistema.
    """
    str_inicio = "    ANO                       CARGA(P.U.DEMANDA MED.)"
    str_fim = "9999"

    def __init__(self):

        super().__init__(BlocoCargaPatamarSubsistemas.str_inicio,
                         "",
                         True)

        self._dados: pd.DataFrame = pd.DataFrame()

    def __eq__(self, o: object):
        if not isinstance(o, BlocoCargaPatamarSubsistemas):
            return False
        bloco: BlocoCargaPatamarSubsistemas = o
        return self._dados.equals(bloco._dados)

    # Override
    def le(self, arq: IO):

        def converte_tabela_em_df() -> pd.DataFrame:
            pats = [n for n in range(1, NUM_PATAMARES + 1)]
            coluna_pats = pats * int(i / NUM_PATAMARES)
            df = pd.DataFrame(tabela)
            df.columns = MESES_DF
            df["Ano"] = anos
            df["Patamar"] = coluna_pats
            df["Subsistema"] = subsistemas
            df = df[["Ano", "Patamar", "Subsistema"] + MESES_DF]
            return df

        # Variáveis auxiliares
        reg_ano = RegistroIn(4)
        reg_pat = RegistroFn(6)
        # Pula as duas primeiras linhas, com cabeçalhos
        arq.readline()
        i = 0
        anos = []
        subsistemas = []
        subsistema_atual = 0
        tabela = np.zeros((MAX_ANOS_ESTUDO * NUM_PATAMARES * len(SUBMERCADOS),
                          len(MESES)))
        while True:
            # Verifica se o arquivo acabou
            linha: str = arq.readline()
            if BlocoCargaPatamarSubsistemas.str_fim == linha.strip():
                tabela = tabela[:i, :]
                self._dados = converte_tabela_em_df()
                break
            # Senão, lê mais uma linha
            # Subsistema
            if linha.strip().isnumeric():
                subsistema_atual = int(linha)
                continue
            subsistemas.append(subsistema_atual)
            # Ano
            if linha[3:7].isnumeric():
                anos.append(reg_ano.le_registro(linha, 3))
            else:
                anos.append(anos[-1])
            # Patamares
            tabela[i, :] = reg_pat.le_linha_tabela(linha,
                                                   8,
                                                   1,
                                                   len(MESES))
            i += 1

    # Override
    def escreve(self, arq: IO):

        def escreve_patamares():
            lin_tab = self._dados.shape[0]
            subsistema_anterior = 0
            i_subsistema = 0
            for i in range(lin_tab):
                linha = ""
                # Subsistema
                subsistema = self._dados.iloc[i, 2]
                if subsistema != subsistema_anterior:
                    linha += str(subsistema).rjust(4)
                    subsistema_anterior = subsistema
                    i_subsistema = 0
                else:
                    # Ano
                    if i_subsistema % NUM_PATAMARES == 0:
                        linha += "   " + str(self._dados.iloc[i, 0]).rjust(4)
                    else:
                        linha += "       "
                    i_subsistema += 1
                    # Patamares de cada mês
                    for j in range(len(MESES)):
                        v = self._dados.iloc[i, j + 3]
                        linha += "{:1.4f}".format(v).rjust(7)
                arq.write(linha + "\n")

        # Escreve cabeçalhos
        cab = ("   XXXX X.XXXX X.XXXX X.XXXX X.XXXX X.XXXX X.XXXX"
               + " X.XXXX X.XXXX X.XXXX X.XXXX X.XXXX X.XXXX" + "\n")
        arq.write(f"{BlocoCargaPatamarSubsistemas.str_inicio}\n")
        arq.write(cab)
        escreve_patamares()
        # Escreve a linha de terminação
        arq.write(f" {BlocoCargaPatamarSubsistemas.str_fim}\n")


class BlocoIntercambioPatamarSubsistemas(Bloco):
    """
    Bloco com a informação de intercâmbio (em p.u.) por patamar de carga
    e por mês/ano de estudo para cada subsistema.
    """
    str_inicio = "                             INTERCAMBIO(P.U.INTERC.MEDIO)"
    str_fim = "9999"

    def __init__(self):

        super().__init__(BlocoIntercambioPatamarSubsistemas.str_inicio,
                         "",
                         True)

        self._dados: pd.DataFrame = pd.DataFrame()

    def __eq__(self, o: object):
        if not isinstance(o, BlocoIntercambioPatamarSubsistemas):
            return False
        bloco: BlocoIntercambioPatamarSubsistemas = o
        return self._dados.equals(bloco._dados)

    # Override
    def le(self, arq: IO):

        def converte_tabela_em_df() -> pd.DataFrame:
            pats = [n for n in range(1, NUM_PATAMARES + 1)]
            coluna_pats = pats * int(i / NUM_PATAMARES)
            df = pd.DataFrame(tabela)
            df.columns = MESES_DF
            df["Ano"] = anos
            df["Patamar"] = coluna_pats
            df["De Subsistema"] = subsistemas_de
            df["Para Subsistema"] = subsistemas_para
            df = df[["Ano", "Patamar", "De Subsistema", "Para Subsistema"]
                    + MESES_DF]
            return df

        # Variáveis auxiliares
        reg_ano = RegistroIn(4)
        reg_pat = RegistroFn(6)
        # Pula a primeira linha, com cabeçalhos
        arq.readline()
        i = 0
        anos = []
        subsistemas_de = []
        subsistemas_para = []
        subsistema_de_atual = 0
        subsistema_para_atual = 0
        tabela = np.zeros((MAX_ANOS_ESTUDO *
                           NUM_PATAMARES *
                           len(SUBMERCADOS) ** 2,
                          len(MESES)))
        while True:
            # Verifica se o arquivo acabou
            linha: str = arq.readline()
            if BlocoIntercambioPatamarSubsistemas.str_fim == linha.strip():
                tabela = tabela[:i, :]
                self._dados = converte_tabela_em_df()
                break
            # Senão, lê mais uma linha
            # Subsistemas de -> para
            if len(linha) < 12:
                subsistema_de_atual = int(linha[1:4].strip())
                subsistema_para_atual = int(linha[5:8].strip())
                continue
            subsistemas_de.append(subsistema_de_atual)
            subsistemas_para.append(subsistema_para_atual)
            # Ano
            if linha[3:7].isnumeric():
                anos.append(reg_ano.le_registro(linha, 3))
            else:
                anos.append(anos[-1])
            # Patamares
            tabela[i, :] = reg_pat.le_linha_tabela(linha,
                                                   8,
                                                   1,
                                                   len(MESES))
            i += 1

    # Override
    def escreve(self, arq: IO):

        def escreve_patamares():
            lin_tab = self._dados.shape[0]
            subsistema_de_anterior = 0
            subsistema_para_anterior = 0
            i_subsistema = 0
            for i in range(lin_tab):
                linha = ""
                # Subsistemas de / para
                subsistema_de = self._dados.iloc[i, 2]
                subsistema_para = self._dados.iloc[i, 3]
                if any([
                        subsistema_de != subsistema_de_anterior,
                        subsistema_para != subsistema_para_anterior
                       ]):
                    linha = (str(subsistema_de).rjust(4) +
                             str(subsistema_para).rjust(4))
                    subsistema_de_anterior = subsistema_de
                    subsistema_para_anterior = subsistema_para
                    i_subsistema = 0
                else:
                    # Ano
                    if i_subsistema % NUM_PATAMARES == 0:
                        linha += "   " + str(self._dados.iloc[i, 0]).rjust(4)
                    else:
                        linha += "       "
                    i_subsistema += 1
                    # Patamares de cada mês
                    for j in range(len(MESES)):
                        v = self._dados.iloc[i, j + 4]
                        linha += "{:1.4f}".format(v).rjust(7)
                arq.write(linha + "\n")

        # Escreve cabeçalhos
        cab = ("   XXXX X.XXXX X.XXXX X.XXXX X.XXXX X.XXXX X.XXXX"
               + " X.XXXX X.XXXX X.XXXX X.XXXX X.XXXX X.XXXX" + "\n")
        arq.write(f"{BlocoIntercambioPatamarSubsistemas.str_inicio}\n")
        arq.write(cab)
        escreve_patamares()
        # Escreve a linha de terminação
        arq.write(f" {BlocoIntercambioPatamarSubsistemas.str_fim}\n")


class LeituraPatamar(Leitura):
    """
    Realiza a leitura do arquivo patamar.dat
    existente em um diretório de entradas do NEWAVE.

    Esta classe contém o conjunto de utilidades para ler
    e interpretar os campos de um arquivo patamar.dat, construindo
    um objeto `Patamar` cujas informações são as mesmas do patamar.dat.

    Este objeto existe para retirar do modelo de dados a complexidade
    de iterar pelas linhas do arquivo, recortar colunas, converter
    tipos de dados, dentre outras tarefas necessárias para a leitura.

    """

    def __init__(self,
                 diretorio: str) -> None:
        super().__init__(diretorio)

    # Override
    def _cria_blocos_leitura(self) -> List[Bloco]:
        return [BlocoDuracaoPatamar(),
                BlocoCargaPatamarSubsistemas(),
                BlocoIntercambioPatamarSubsistemas()]



from inewave._utils.arquivo import Arquivo
from inewave._utils.dadosarquivo import DadosArquivo
from inewave._utils.escrita import Escrita
from inewave.newave.modelos.re import LeituraRE

import pandas as pd  # type: ignore


class RE(Arquivo):
    """
    Armazena os dados de entrada do NEWAVE referentes às restrições
    elétricas existentes.

    **Parâmetros**

    """

    def __init__(self,
                 dados: DadosArquivo) -> None:
        super().__init__(dados)

    # Override
    @classmethod
    def le_arquivo(cls,
                   diretorio: str,
                   nome_arquivo="re.dat") -> 'RE':
        """
        """
        leitor = LeituraRE(diretorio)
        r = leitor.le_arquivo(nome_arquivo)
        return cls(r)

    def escreve_arquivo(self,
                        diretorio: str,
                        nome_arquivo="re.dat"):
        """
        """
        escritor = Escrita(diretorio)
        escritor.escreve_arquivo(self._dados,
                                 nome_arquivo)

    @property
    def usinas_conjuntos(self) -> pd.DataFrame:
        """
        Tabela com os conjuntos de usinas com restrições elétricas.

        **Retorna**

        `pd.DataFrame`

        **Sobre**
        """
        return self._blocos[0].dados

    @usinas_conjuntos.setter
    def usinas_conjuntos(self, conjuntos: pd.DataFrame):
        self._blocos[0].dados = conjuntos

    @property
    def restricoes(self) -> pd.DataFrame:
        """
        Tabela com as configurações das restrições elétricas.

        **Retorna**

        `pd.DataFrame`

        **Sobre**
        """
        return self._blocos[1].dados

    @restricoes.setter
    def restricoes(self, config: pd.DataFrame):
        self._blocos[1].dados = config



from inewave.config import MAX_RES
from inewave._utils.bloco import Bloco
from inewave._utils.leitura import Leitura
from inewave._utils.registros import RegistroFn, RegistroIn

from typing import List, IO
import numpy as np  # type: ignore
import pandas as pd  # type: ignore


class BlocoUsinasConjuntoRE(Bloco):
    """
    Bloco com informações das usinas pertencentes a cada conjunto
    de restrições elétricas por conjunto de RE.
    """
    str_inicio = "RES   USINAS PERTENCENTES AO CONJUNTO"
    str_fim = "999"

    def __init__(self):

        super().__init__(BlocoUsinasConjuntoRE.str_inicio,
                         "",
                         True)

        self._dados: pd.DataFrame = pd.DataFrame()

    def __eq__(self, o: object):
        if not isinstance(o, BlocoUsinasConjuntoRE):
            return False
        bloco: BlocoUsinasConjuntoRE = o
        return self._dados.equals(bloco._dados)

    # Override
    def le(self, arq: IO):

        def converte_tabela_em_df() -> pd.DataFrame:
            df = pd.DataFrame(tabela)
            cols_usinas = [f"Usina {i}" for i in range(1, 11)]
            df.columns = cols_usinas
            df["Conjunto"] = conjuntos
            df = df[["Conjunto"] + cols_usinas]
            return df

        # Variáveis auxiliares
        reg_conjunto = RegistroIn(3)
        # Pula uma linha, com cabeçalhos
        arq.readline()
        i = 0
        conjuntos = []
        tabela = np.zeros((MAX_RES, 10), dtype=np.int64)
        while True:
            # Verifica se o arquivo acabou
            linha: str = arq.readline()
            if (BlocoUsinasConjuntoRE.str_fim
               == linha.strip()):
                tabela = tabela[:i, :]
                self._dados = converte_tabela_em_df()
                break
            conjuntos.append(reg_conjunto.le_registro(linha, 0))
            col_i = 6
            # Lê a linha conferindo quantas colunas tem
            linha_tabela: List[int] = []
            for j in range(1, 11):
                col_f = col_i + 3
                if col_f >= len(linha):
                    break
                linha_tabela.append(int(linha[col_i:col_f]))
                col_i = col_f + 1
            for j, usi in enumerate(linha_tabela):
                tabela[i, j] = usi
            i += 1

    # Override
    def escreve(self, arq: IO):

        def escreve_conjuntos():
            lin_tab = self._dados.shape[0]
            for i in range(lin_tab):
                linha = f"{str(int(self._dados.iloc[i, 0])).rjust(3)}  "
                for j in range(1, 11):
                    usi = self._dados.iloc[i, j]
                    if usi == 0:
                        continue
                    linha += f" {str(self._dados.iloc[i, j]).rjust(3)}"
                arq.write(linha + "\n")

        # Escreve cabeçalhos
        arq.write(f"{BlocoUsinasConjuntoRE.str_inicio}\n")
        arq.write("XXX   XXX XXX XXX XXX XXX XXX XXX XXX XXX XXX\n")
        escreve_conjuntos()
        # Escreve a linha de terminação
        arq.write(f"{BlocoUsinasConjuntoRE.str_fim}\n")


class BlocoConfiguracaoRestricoesRE(Bloco):
    """
    Bloco com informações de configuração das restrições elétricas
    para cada conjunto de usinas.
    """
    str_inicio = "RES MM/AAAA MM/AAAA P       RESTRICAO"
    str_fim = "999"

    def __init__(self):

        super().__init__(BlocoConfiguracaoRestricoesRE.str_inicio,
                         "",
                         True)

        self._dados: pd.DataFrame = pd.DataFrame()

    def __eq__(self, o: object):
        if not isinstance(o, BlocoConfiguracaoRestricoesRE):
            return False
        bloco: BlocoConfiguracaoRestricoesRE = o
        return self._dados.equals(bloco._dados)

    # Override
    def le(self, arq: IO):

        def converte_tabela_em_df() -> pd.DataFrame:
            df = pd.DataFrame(tabela)
            df.columns = ["Conjunto", "Mês Início", "Ano Início",
                          "Mês Fim", "Ano Fim", "Flag P"]
            df["Restrição"] = restricoes
            df["Motivo"] = motivos
            return df

        # Variáveis auxiliares
        reg_conjunto = RegistroIn(3)
        reg_mes = RegistroIn(2)
        reg_ano = RegistroIn(4)
        reg_flag = RegistroIn(1)
        reg_resticao = RegistroFn(15)
        # Pula uma linha, com cabeçalhos
        arq.readline()
        i = 0
        tabela = np.zeros((MAX_RES, 6), dtype=np.int64)
        restricoes = []
        motivos = []
        while True:
            # Verifica se o arquivo acabou
            linha: str = arq.readline()
            if (BlocoUsinasConjuntoRE.str_fim
               == linha.strip()):
                tabela = tabela[:i, :]
                self._dados = converte_tabela_em_df()
                break
            tabela[i, 0] = reg_conjunto.le_registro(linha, 0)
            tabela[i, 1] = reg_mes.le_registro(linha, 4)
            tabela[i, 2] = reg_ano.le_registro(linha, 7)
            tabela[i, 3] = reg_mes.le_registro(linha, 12)
            tabela[i, 4] = reg_ano.le_registro(linha, 15)
            tabela[i, 5] = reg_flag.le_registro(linha, 20)
            restricoes.append(reg_resticao.le_registro(linha, 22))
            motivos.append(linha[38:].strip())
            i += 1

    # Override
    def escreve(self, arq: IO):

        def escreve_restricoes():
            lin_tab = self._dados.shape[0]
            for i in range(lin_tab):
                linha = f"{str(int(self._dados.iloc[i, 0])).rjust(3)}"
                linha += f" {str(int(self._dados.iloc[i, 1])).rjust(2)}"
                linha += f" {str(int(self._dados.iloc[i, 2])).rjust(4)}"
                linha += f" {str(int(self._dados.iloc[i, 3])).rjust(2)}"
                linha += f" {str(int(self._dados.iloc[i, 4])).rjust(4)}"
                linha += f" {str(int(self._dados.iloc[i, 5]))}"
                linha += " " + f"{self._dados.iloc[i, 6]:12.2f}".rjust(15)
                linha += f" {self._dados.iloc[i, 7]}"
                arq.write(linha + "\n")

        # Escreve cabeçalhos
        arq.write(f"{BlocoConfiguracaoRestricoesRE.str_inicio}\n")
        arq.write("XXX XX XXXX XX XXXX X XXXXXXXXXXXXXXX\n")
        escreve_restricoes()
        # Escreve a linha de terminação
        arq.write(f"{BlocoConfiguracaoRestricoesRE.str_fim}\n")


class LeituraRE(Leitura):
    """
    Realiza a leitura do arquivo `re.dat`
    existente em um diretório de entradas do NEWAVE.

    Esta classe contém o conjunto de utilidades para ler
    e interpretar os campos de um arquivo `re.dat`, construindo
    um objeto `RE` cujas informações são as mesmas do re.dat.

    Este objeto existe para retirar do modelo de dados a complexidade
    de iterar pelas linhas do arquivo, recortar colunas, converter
    tipos de dados, dentre outras tarefas necessárias para a leitura.
    """

    def __init__(self,
                 diretorio: str):
        super().__init__(diretorio)

    # Override
    def _cria_blocos_leitura(self) -> List[Bloco]:
        """
        Cria a lista de blocos a serem lidos no arquivo re.dat.
        """
        return [BlocoUsinasConjuntoRE(),
                BlocoConfiguracaoRestricoesRE()]



# Imports do próprio módulo
from inewave.config import MAX_ITERS, REES
from inewave._utils.bloco import Bloco
from inewave._utils.leitura import Leitura
# Imports de módulos externos
import numpy as np  # type: ignore
from typing import IO, List


class RegistroNwlistcf:
    """
    Armazena as informações de um registro da FCF do NEWAVE.

    ** Parâmetros **

    - ireg: `int`
    - rhs:  `float`
    - tabela: `np.ndarray`

    """

    __slots__ = ["ireg",
                 "rhs",
                 "tabela"]

    def __init__(self,
                 ireg: int,
                 rhs: float,
                 tabela: np.ndarray):
        self.ireg = ireg
        self.rhs = rhs
        self.tabela = tabela

    @classmethod
    def le_registro(cls, primeira_linha: str, arq: IO) -> 'RegistroNwlistcf':
        """
        """
        primeira = True
        n_rees = len(REES)
        n_cols_tabela = 16
        ireg = 0
        rhs = 0.0
        tabela = np.zeros((n_rees, n_cols_tabela))
        for i in range(n_rees):
            if primeira:
                primeira = False
                linha = primeira_linha
                # Extrai os campos específicos da primeira linha
                ireg = int(linha[2:10])
                rhs = float(linha[15:30])
            # Preenche a tabela com os dados do registro
            else:
                linha = arq.readline()
            ree = int(linha[11:14])
            ci = 31
            nc = 17
            for j in range(n_cols_tabela):
                cf = ci + nc
                num_str = linha[ci:cf]
                valor = 0.0 if not num_str.isnumeric() else float(num_str)
                tabela[ree-1, j] = valor
                ci = cf + 1

        return cls(ireg, rhs, tabela)

    def __eq__(self, o: object) -> bool:
        """
        A igualdade entre RegistroNwlistcf avalia todos os
        valores.
        """
        if not isinstance(o, RegistroNwlistcf):
            return False
        reg: RegistroNwlistcf = o
        eq_ireg = self.ireg == reg.ireg
        eq_rhs = self.rhs == reg.rhs
        eq_tab = np.array_equal(self.tabela, reg.tabela)

        return all([eq_ireg,
                    eq_rhs,
                    eq_tab])


class BlocoPeriodoNwlistcf(Bloco):
    """
    Bloco com informações dos cortes de um período,
    existentes no arquivo `nwlistcf.rel` do NWLISTCF.
    """
    str_inicio = "  PERIODO:  "

    def __init__(self):

        super().__init__(BlocoPeriodoNwlistcf.str_inicio,
                         "",
                         False)

        self._dados: List[RegistroNwlistcf] = []

    def __eq__(self, o: object):
        if not isinstance(o, BlocoPeriodoNwlistcf):
            return False
        bloco: BlocoPeriodoNwlistcf = o
        return all([d1 == d2
                    for d1, d2 in zip(self._dados, bloco._dados)])

    # Override
    def le(self, arq: IO):
        # Salta duas linhas para acessar a tabela
        arq.readline()
        arq.readline()
        while True:
            # Verifica se a próxima linha é o início do
            # próximo período ou é vazia
            linha = arq.readline()
            if (BlocoPeriodoNwlistcf.str_inicio in linha
                    or len(linha) < 2):
                break
            # Senão, lê mais um registro
            reg = RegistroNwlistcf.le_registro(linha, arq)
            self._dados.append(reg)

        return linha

    # Override
    def escreve(self, arq: IO):
        pass


class LeituraNwlistcf(Leitura):
    """
    Realiza a leitura do arquivo nwlistcf.rel,
    existente em um diretório de saídas do NWLISTCF.

    Esta classe contém o conjunto de utilidades para ler
    e interpretar os campos do arquivo nwlistcf.rel, construindo um
    objeto `Nwlistcf` cujas informações são as mesmas do arquivo.

    Este objeto existe para retirar do modelo de dados a complexidade
    de iterar pelas linhas do arquivo, recortar colunas, converter
    tipos de dados, dentre outras tarefas necessárias para a leitura.

    """
    str_inicio_periodo = "  PERIODO:      "

    def __init__(self,
                 diretorio: str) -> None:
        super().__init__(diretorio)

    def _cria_blocos_leitura(self) -> List[Bloco]:
        return [BlocoPeriodoNwlistcf()
                for _ in range(MAX_ITERS)]



from inewave.nwlistcf.estados import Estados

est = Estados.le_arquivo("./tests/_arquivos")


def test_leitura():
    assert len(est.registros) > 0


def test_eq_estados():
    est2 = Estados.le_arquivo("./tests/_arquivos")
    assert est == est2


# def test_neq_estados():
#     leitor2 = LeituraEstados("./tests/_arquivos")
#     leitor2.le_arquivo()
#     leitor2.estados.registros[5] = {}
#     assert leitor2.estados != leitor.estados



from inewave.newave.modelos.confhd import BlocoConfUHE
from inewave.newave.modelos.confhd import LeituraConfhd
from inewave._utils.dadosarquivo import DadosArquivo
from inewave._utils.arquivo import Arquivo
from inewave._utils.escrita import Escrita

import pandas as pd  # type: ignore


class Confhd(Arquivo):
    """
    Armazena os dados de entrada do NEWAVE referentes às
    configurações das usinas hidrelétricas.

    Esta classe lida com informações de entrada fornecidas ao NEWAVE e
    que podem ser modificadas através do arquivo `modif.dat`.

    """
    def __init__(self,
                 dados: DadosArquivo):
        super().__init__(dados)
        # Interpreta o resultado da leitura
        val = True
        msg = "Erro na criação de Confhd: "
        if len(dados.blocos) == 1:
            bloco = dados.blocos[0]
            if isinstance(bloco, BlocoConfUHE):
                self.__bloco = bloco
            else:
                msg += (f"O bloco deve ser do tipo {BlocoConfUHE}, " +
                        f"mas foi fornecido do tipo {type(bloco)}")
                val = False
        else:
            msg += "Deve ser fornecido exatamente 1 bloco para Confhd"
            val = False
        if not val:
            raise TypeError(msg)

    @classmethod
    def le_arquivo(cls,
                   diretorio: str,
                   nome_arquivo="confhd.dat") -> 'Confhd':
        """
        """
        leitor = LeituraConfhd(diretorio)
        r = leitor.le_arquivo(nome_arquivo)
        return cls(r)

    def escreve_arquivo(self,
                        diretorio: str,
                        nome_arquivo="confhd.dat"):
        """
        """
        escritor = Escrita(diretorio)
        escritor.escreve_arquivo(self._dados,
                                 nome_arquivo)

    @property
    def usinas(self) -> pd.DataFrame:
        return self.__bloco.dados



from .dadosarquivo import DadosArquivo
from .bloco import Bloco

import os
from typing import IO, List, Dict
from traceback import print_exc


class Escrita:
    """
    Classe com utilidades gerais para a escrita de arquivos
    do NEWAVE.
    """
    def __init__(self,
                 diretorio: str):
        self._diretorio = diretorio

    def _escreve_blocos_e_linhas(self,
                                 arq: IO,
                                 blocos: List[Bloco],
                                 linhas: Dict[int, str]):

        ordem_blocos = [b._ordem for b in blocos]
        ordem_linhas = list(linhas.keys())
        num_itens = max(ordem_blocos + ordem_linhas) + 1
        for i in range(num_itens):
            if i in ordem_blocos:
                blocos[ordem_blocos.index(i)].escreve(arq)
            elif i in ordem_linhas:
                arq.write(linhas[i])

    def escreve_arquivo(self,
                        dados: DadosArquivo,
                        nome_arquivo: str):
        """
        """
        try:
            if not os.path.exists(self._diretorio):
                os.makedirs(self._diretorio)
            caminho = os.path.join(self._diretorio, nome_arquivo)
            with open(caminho, "w") as arq:
                self._escreve_blocos_e_linhas(arq,
                                              dados.blocos,
                                              dados.linhas_fora_blocos)
        except Exception:
            print_exc()



# Rotinas de testes associadas ao arquivo eafpast.dat do NEWAVE
from inewave.newave.eafpast import EafPast


eaf = EafPast.le_arquivo("tests/_arquivos")


def test_leitura():
    assert len(eaf.tendencia) > 0


def test_escrita_e_leitura():
    eaf.escreve_arquivo("tests/_saidas")
    eaf2 = EafPast.le_arquivo("tests/_saidas")
    assert eaf == eaf2


def test_neq_vazpast():
    eaf2 = EafPast.le_arquivo("tests/_arquivos")
    eaf2.tendencia.iloc[0, 0] = -1
    assert eaf != eaf2




from typing import Dict, List

from.bloco import Bloco


class DadosArquivo:
    """
    """
    def __init__(self,
                 blocos: List[Bloco],
                 linhas_fora_blocos: Dict[int, str]) -> None:
        self.__blocos = blocos
        self.__linhas_fora_blocos = linhas_fora_blocos
        pass

    @property
    def blocos(self) -> List[Bloco]:
        return self.__blocos

    @property
    def linhas_fora_blocos(self) -> Dict[int, str]:
        return self.__linhas_fora_blocos



# Rotinas de testes associadas ao arquivo vazpast.dat do NEWAVE
from inewave.newave.vazpast import VazPast


vaz = VazPast.le_arquivo("tests/_arquivos")


def test_leitura():
    assert len(vaz.tendencia) > 0


def test_escrita_e_leitura():
    vaz.escreve_arquivo("tests/_saidas")
    vaz2 = VazPast.le_arquivo("tests/_saidas")
    assert vaz == vaz2


def test_neq_vazpast():
    vaz2 = VazPast.le_arquivo("tests/_arquivos")
    vaz2.tendencia.iloc[0, 0] = -1
    assert vaz != vaz2



from inewave._utils.arquivo import Arquivo
from inewave._utils.dadosarquivo import DadosArquivo
from inewave._utils.escrita import Escrita
from inewave.newave.modelos.adterm import LeituraAdTerm

import pandas as pd  # type: ignore


class AdTerm(Arquivo):
    """
    Armazena os dados de entrada do NEWAVE referentes às térmicas de
    despacho antecipado disponíveis.

    **Parâmetros**

    """

    def __init__(self,
                 dados: DadosArquivo) -> None:
        super().__init__(dados)

    # Override
    @classmethod
    def le_arquivo(cls,
                   diretorio: str,
                   nome_arquivo="adterm.dat") -> 'AdTerm':
        """
        """
        leitor = LeituraAdTerm(diretorio)
        r = leitor.le_arquivo(nome_arquivo)
        return cls(r)

    def escreve_arquivo(self,
                        diretorio: str,
                        nome_arquivo="adterm.dat"):
        """
        """
        escritor = Escrita(diretorio)
        escritor.escreve_arquivo(self._dados,
                                 nome_arquivo)

    @property
    def despachos(self) -> pd.DataFrame:
        """
        Despachos antecipados das térmicas GNL.

        **Retorna**

        `pd.DataFrame`

        """
        return self._blocos[0].dados

    @despachos.setter
    def despachos(self, term: pd.DataFrame):
        self._blocos[0].dados = term



# Rotinas de testes associadas ao arquivo earmfpm00x.out do NWLISTOP
from inewave.nwlistop.earmfpm00 import EarmfpM00

sub_teste = "SUDESTE"
earm = EarmfpM00.le_arquivo("tests/_arquivos", "earmfpm00test.out")


def test_leitura():
    assert sub_teste in earm.submercado


def test_eq_earmfpm00():
    earm2 = EarmfpM00.le_arquivo("tests/_arquivos", "earmfpm00test.out")
    assert earm == earm2


def test_neq_earmfpm00():
    earm2 = EarmfpM00.le_arquivo("tests/_arquivos", "earmfpm00test.out")
    earm2.energias.iloc[0, 0] = -1
    assert earm != earm2



# Imports do próprio módulo
from inewave._utils.registros import RegistroAn, RegistroFn, RegistroIn
from inewave._utils.bloco import Bloco
from inewave._utils.leitura import Leitura
from inewave.config import MAX_UHES, MESES, MESES_DF
# Imports de módulos externos
import numpy as np  # type: ignore
import pandas as pd  # type: ignore
from typing import IO, List


class BlocoVazPast(Bloco):
    """
    Bloco de informações de vazões passadas
    por usina, existentes no arquivo `vazpast.dat`
    do NEWAVE.
    """
    str_inicio = ""

    def __init__(self):

        super().__init__(BlocoVazPast.str_inicio,
                         "",
                         True)

        self._dados = [0, 0, pd.DataFrame]

    def __eq__(self, o: object):
        if not isinstance(o, BlocoVazPast):
            return False
        bloco: BlocoVazPast = o
        return all([
                    self._dados[0] == bloco._dados[0],
                    self._dados[1] == bloco._dados[1],
                    self._dados[2].equals(bloco._dados[2])
                   ])

    # Override
    def le(self, arq: IO):
        # Pula as duas primeiras linhas, com cabeçalhos
        arq.readline()
        # Extrai o mês e ano do estudo
        linha = arq.readline()

        str_mes_plan, str_ano_plan = linha.split("ANOPLAN")
        self._dados[0] = int(str_mes_plan.split("=")[1].strip())
        self._dados[1] = int(str_ano_plan.split("=")[1].strip())
        # Variáveis auxiliares
        reg_posto = RegistroIn(3)
        reg_nome = RegistroAn(12)
        reg_vaz = RegistroFn(8)
        i = 0
        postos: List[int] = []
        nomes: List[str] = []
        tabela = np.zeros((MAX_UHES, len(MESES)))
        while True:
            # Verifica se o arquivo acabou
            linha = arq.readline()
            if len(linha) < 3:
                tabela = tabela[:i, :]
                self._dados[2] = pd.DataFrame(tabela)
                self._dados[2].columns = MESES_DF
                self._dados[2]["Posto"] = postos
                self._dados[2]["Usina"] = nomes
                self._dados[2] = self._dados[2][["Posto", "Usina"] + MESES_DF]
                break
            # Senão, lê mais uma linha
            postos.append(reg_posto.le_registro(linha, 2))
            nomes.append(reg_nome.le_registro(linha, 6))
            tabela[i, :] = reg_vaz.le_linha_tabela(linha,
                                                   20,
                                                   2,
                                                   len(MESES))
            i += 1

    # Override
    def escreve(self, arq: IO):

        def escreve_desvios():
            lin_tab = self._dados[2].shape[0]
            for i in range(lin_tab):
                linha = " "
                # Posto
                linha += str(self._dados[2].iloc[i, 0]).rjust(4) + " "
                # Nome
                linha += str(self._dados[2].iloc[i, 1]).ljust(12)
                # Vazoes de cada mês
                for j in range(len(MESES)):
                    v = self._dados[2].iloc[i, 2 + j]
                    linha += " " + "{:6.2f}".format(float(v)).rjust(9)
                arq.write(linha + "\n")

        # Escreve cabeçalhos
        dummy = ("********* DUMMY" + "\n")
        arq.write(dummy)
        arq.write(dummy)
        ano_mes = " "
        ano_mes += f"MESPLAN={str(self._dados[0]).rjust(13)}"
        ano_mes += f"  ANOPLAN={str(self._dados[1]).rjust(13)}"
        arq.write(ano_mes + "\n")
        escreve_desvios()


class LeituraVazPast(Leitura):
    """
    Realiza a leitura do arquivo `vazpast.dat`
    existente em um diretório de entradas do NEWAVE.

    Esta classe contém o conjunto de utilidades para ler
    e interpretar os campos de um arquivo `vazpast.dat`, construindo
    um objeto `VazPast` cujas informações são as mesmas do `vazpast.dat`.

    Este objeto existe para retirar do modelo de dados a complexidade
    de iterar pelas linhas do arquivo, recortar colunas, converter
    tipos de dados, dentre outras tarefas necessárias para a leitura.

    """

    def __init__(self,
                 diretorio: str) -> None:
        super().__init__(diretorio)

    def _cria_blocos_leitura(self) -> List[Bloco]:
        return [BlocoVazPast()]



from inewave.config import SUBMERCADOS, MESES, MESES_DF, MAX_ANOS_ESTUDO
from inewave._utils.bloco import Bloco
from inewave._utils.leitura import Leitura
from inewave._utils.registros import RegistroAn, RegistroIn, RegistroFn

from typing import List, IO
import numpy as np  # type: ignore
import pandas as pd  # type: ignore


class BlocoCustoDeficitSistema(Bloco):
    """
    Bloco com informações sobre o custo de déficit por
    patamar de déficit e o número de patamares de déficit.
    """
    str_inicio = " NUMERO DE PATAMARES DE DEFICIT"
    str_fim = " 999"

    def __init__(self):

        super().__init__(BlocoCustoDeficitSistema.str_inicio,
                         "",
                         True)

        self._dados = [0, pd.DataFrame()]

    def __eq__(self, o: object):
        if not isinstance(o, BlocoCustoDeficitSistema):
            return False
        bloco: BlocoCustoDeficitSistema = o
        return all([
                    self._dados[0] == bloco._dados[0],
                    self._dados[1].equals(bloco._dados[1])
                   ])

    # Override
    def le(self, arq: IO):

        def converte_tabela_em_df() -> pd.DataFrame:
            df = pd.DataFrame(tabela)
            pats = [n for n in range(1, num_pat + 1)]
            cols_custo_pats = [f"Custo Pat. {p}" for p in pats]
            cols_corte_pats = [f"Corte Pat. {p}" for p in pats]
            df.columns = cols_custo_pats + cols_corte_pats
            df["Num. Subsistema"] = num_subsistemas
            df["Nome"] = nomes_subsistemas
            df["Fictício"] = ficticios
            df = df[["Num. Subsistema", "Nome", "Fictício"] +
                    cols_custo_pats +
                    cols_corte_pats]
            return df

        # Variáveis auxiliares
        reg_num = RegistroIn(3)
        reg_nome_subsis = RegistroAn(10)
        reg_flag_ficticio = RegistroIn(1)
        reg_custos = RegistroFn(7)
        reg_pu_corte = RegistroFn(5)
        # Pula a primeira linha, com cabeçalhos
        arq.readline()
        # Lê o número de patamares de déficit
        num_pat = int(reg_num.le_registro(arq.readline(), 1))
        self._dados[0] = num_pat
        # Pula três linhas
        for _ in range(3):
            arq.readline()
        num_subsistemas = []
        nomes_subsistemas = []
        ficticios = []
        tabela = np.zeros((len(SUBMERCADOS) + 1,
                           2 * num_pat))
        for i in range(len(SUBMERCADOS)):
            linha: str = arq.readline()
            num_subsistemas.append(reg_num.le_registro(linha, 1))
            nomes_subsistemas.append(reg_nome_subsis.le_registro(linha, 5))
            ficticios.append(bool(reg_flag_ficticio.le_registro(linha, 17)))
            tabela[i, :num_pat] = reg_custos.le_linha_tabela(linha,
                                                             19,
                                                             1,
                                                             num_pat)
            tabela[i, num_pat:] = reg_pu_corte.le_linha_tabela(linha,
                                                               51,
                                                               1,
                                                               num_pat)
        # Adiciona a linha do fictício
        linha = arq.readline()
        num_subsistemas.append(reg_num.le_registro(linha, 1))
        nomes_subsistemas.append(reg_nome_subsis.le_registro(linha, 5))
        ficticios.append(bool(reg_flag_ficticio.le_registro(linha, 17)))

        self._dados[1] = converte_tabela_em_df()

        # Confirma se terminou
        if BlocoCustoDeficitSistema.str_fim not in arq.readline():
            raise ValueError("Arquivo sistema.dat inválido")

    # Override
    def escreve(self, arq: IO):

        def escreve_custos():
            npat = self._dados[0]
            tabela = self._dados[1]
            lin_tab = tabela.shape[0]
            for i in range(lin_tab):
                linha = ""
                num = tabela.iloc[i, 0]
                nome = tabela.iloc[i, 1]
                fict = int(tabela.iloc[i, 2])
                linha += f" {str(num).ljust(3)}"
                linha += f" {str(nome).ljust(10)}"
                linha += f"  {str(fict)}"
                if num < len(SUBMERCADOS) + 1:
                    for j in range(npat):
                        custo = tabela.iloc[i, j + 3]
                        linha += " " + f"{custo:4.2f}".rjust(7)
                    for j in range(npat):
                        corte = tabela.iloc[i, j + npat + 3]
                        linha += " " + f"{corte:1.3f}".rjust(5)
                arq.write(linha + "\n")

        # Escreve cabeçalhos
        arq.write(f"{BlocoCustoDeficitSistema.str_inicio}\n")
        arq.write(" XXX\n")
        arq.write(f" {str(self._dados[0]).rjust(3)}\n")
        arq.write(" CUSTO DO DEFICIT\n")
        arq.write(" NUM|NOME SSIS.|    CUSTO DE DEFICIT POR PATAMAR" +
                  "  | P.U. CORTE POR PATAMAR|\n")
        arq.write(" XXX|XXXXXXXXXX| F|XXXX.XX XXXX.XX XXXX.XX XXXX.XX|" +
                  "X.XXX X.XXX X.XXX X.XXX|\n")
        escreve_custos()
        # Escreve a linha de terminação
        arq.write(f"{BlocoCustoDeficitSistema.str_fim}\n")


class BlocoIntercambioSistema(Bloco):
    """
    Bloco com a informação de intercâmbio
    por mês/ano de estudo para cada subsistema.
    """
    str_inicio = " LIMITES DE INTERCAMBIO"
    str_fim = "999"

    def __init__(self):

        super().__init__(BlocoIntercambioSistema.str_inicio,
                         "",
                         True)

        self._dados: pd.DataFrame = pd.DataFrame()

    def __eq__(self, o: object):
        if not isinstance(o, BlocoIntercambioSistema):
            return False
        bloco: BlocoIntercambioSistema = o
        return self._dados.equals(bloco._dados)

    # Override
    def le(self, arq: IO):

        def converte_tabela_em_df() -> pd.DataFrame:
            df = pd.DataFrame(tabela)
            df.columns = MESES_DF
            df["Ano"] = anos
            df["De Subsistema"] = subsistemas_de
            df["Para Subsistema"] = subsistemas_para
            df = df[["Ano", "De Subsistema", "Para Subsistema"]
                    + MESES_DF]
            return df

        # Variáveis auxiliares
        reg_ano = RegistroIn(4)
        reg_lim = RegistroFn(7)
        # Pula duas linhas, com cabeçalhos
        arq.readline()
        arq.readline()
        i = 0
        anos = []
        subsistemas_de = []
        subsistemas_para = []
        subsis_de_atual = 0
        subsis_para_atual = 0
        tabela = np.zeros((MAX_ANOS_ESTUDO *
                           2 *
                           len(SUBMERCADOS) ** 2,
                          len(MESES)))
        while True:
            # Verifica se o arquivo acabou
            linha: str = arq.readline()
            if BlocoIntercambioSistema.str_fim == linha.strip():
                tabela = tabela[:i, :]
                self._dados = converte_tabela_em_df()
                break
            # Senão, lê mais uma linha
            # Subsistemas de -> para
            if 5 < len(linha.strip()) < 30:
                subsis_de_atual = int(linha[1:4].strip())
                subsis_para_atual = int(linha[5:8].strip())
                continue
            elif len(linha.strip()) < 5:
                subsis_para_atual, subsis_de_atual = (subsis_de_atual,
                                                      subsis_para_atual)
            else:
                subsistemas_de.append(subsis_de_atual)
                subsistemas_para.append(subsis_para_atual)
                # Ano
                anos.append(reg_ano.le_registro(linha, 0))
                # Limites
                tabela[i, :] = reg_lim.le_linha_tabela(linha,
                                                       7,
                                                       1,
                                                       len(MESES))
                i += 1

    # Override
    def escreve(self, arq: IO):

        def escreve_limites():
            lin_tab = self._dados.shape[0]
            subsistema_de_anterior = 0
            subsistema_para_anterior = 0
            for i in range(lin_tab):
                linha = ""
                # Subsistemas de / para
                subsistema_de = self._dados.iloc[i, 1]
                subsistema_para = self._dados.iloc[i, 2]
                if any([
                        subsistema_de != subsistema_de_anterior,
                        subsistema_para != subsistema_para_anterior
                       ]):
                    if not all([
                                subsistema_de == subsistema_para_anterior,
                                subsistema_para == subsistema_de_anterior
                               ]):
                        linha = (str(subsistema_de).rjust(4) +
                                 str(subsistema_para).rjust(4) +
                                 "               0")
                    arq.write(linha + "\n")
                    subsistema_de_anterior = subsistema_de
                    subsistema_para_anterior = subsistema_para

                # Patamares de cada mês
                linha = f"{self._dados.iloc[i, 0]}   "
                for j in range(len(MESES)):
                    v = self._dados.iloc[i, j + 3]
                    linha += " {:6.0f}".format(v).rjust(7)
                arq.write(linha + "\n")

        # Escreve cabeçalhos
        arq.write(f"{BlocoIntercambioSistema.str_inicio}\n")
        arq.write(" A   B   A->B    B->A\n")
        cab = (" XXX XXX XJAN. XXXFEV. XXXMAR. XXXABR. XXXMAI." +
               " XXXJUN. XXXJUL. XXXAGO. XXXSET. XXXOUT. XXXNOV. XXXDEZ.\n")
        arq.write(cab)
        escreve_limites()
        # Escreve a linha de terminação
        arq.write(f" {BlocoIntercambioSistema.str_fim}\n")


class BlocoMercadoEnergiaSistema(Bloco):
    """
    Bloco com a informação de mercado de energia
    por mês/ano de estudo para cada subsistema.
    """
    str_inicio = " MERCADO DE ENERGIA TOTAL"
    str_fim = "999"

    def __init__(self):

        super().__init__(BlocoMercadoEnergiaSistema.str_inicio,
                         "",
                         True)

        self._dados: pd.DataFrame = pd.DataFrame()

    def __eq__(self, o: object):
        if not isinstance(o, BlocoMercadoEnergiaSistema):
            return False
        bloco: BlocoMercadoEnergiaSistema = o
        return self._dados.equals(bloco._dados)

    # Override
    def le(self, arq: IO):

        def converte_tabela_em_df() -> pd.DataFrame:
            df = pd.DataFrame(tabela)
            df.columns = MESES_DF
            df["Ano"] = anos
            df["Subsistema"] = subsistema
            df = df[["Ano", "Subsistema"] + MESES_DF]
            return df

        # Variáveis auxiliares
        reg_subsis = RegistroIn(3)
        reg_ano = RegistroAn(4)
        reg_merc = RegistroFn(7)
        # Pula duas linhas, com cabeçalhos
        arq.readline()
        arq.readline()
        i = 0
        anos = []
        subsistema = []
        subsistema_atual = 0
        tabela = np.zeros((MAX_ANOS_ESTUDO * len(SUBMERCADOS),
                          len(MESES)))
        while True:
            # Verifica se o arquivo acabou
            linha: str = arq.readline()
            if BlocoMercadoEnergiaSistema.str_fim == linha.strip():
                tabela = tabela[:i, :]
                self._dados = converte_tabela_em_df()
                break
            # Senão, lê mais uma linha
            if len(linha.strip()) < 5:
                subsistema_atual = reg_subsis.le_registro(linha, 1)
            else:
                # Ano
                anos.append(reg_ano.le_registro(linha, 0))
                # Subsistema
                subsistema.append(subsistema_atual)
                # Limites
                tabela[i, :] = reg_merc.le_linha_tabela(linha,
                                                        7,
                                                        1,
                                                        len(MESES))
                i += 1

    # Override
    def escreve(self, arq: IO):

        def escreve_mercados():
            lin_tab = self._dados.shape[0]
            subsistema_anterior = 0
            for i in range(lin_tab):
                linha = ""
                # Subsistema
                subsistema = self._dados.iloc[i, 1]
                if subsistema != subsistema_anterior:
                    arq.write(" " + str(subsistema).rjust(3) + "\n")
                    subsistema_anterior = subsistema
                # Mercados de cada mês
                linha = f"{self._dados.iloc[i, 0].ljust(4)}   "
                for j in range(len(MESES)):
                    v = self._dados.iloc[i, j + 2]
                    linha += " {:6.0f}".format(v).rjust(7)
                arq.write(linha + "\n")

        # Escreve cabeçalhos
        arq.write(f"{BlocoMercadoEnergiaSistema.str_inicio}\n")
        arq.write(" XXX\n")
        cab = ("       XXXJAN. XXXFEV. XXXMAR. XXXABR. XXXMAI. XXXJUN." +
               " XXXJUL. XXXAGO. XXXSET. XXXOUT. XXXNOV. XXXDEZ.\n")
        arq.write(cab)
        escreve_mercados()
        # Escreve a linha de terminação
        arq.write(f" {BlocoMercadoEnergiaSistema.str_fim}\n")


class BlocoGeracaoUsinasNaoSimuladasSistema(Bloco):
    """
    Bloco com a informação de geração das usinas não simuladas
    por mês/ano de estudo para cada subsistema.
    """
    str_inicio = " GERACAO DE USINAS NAO SIMULADAS"
    str_fim = "999"

    def __init__(self):

        super().__init__(BlocoGeracaoUsinasNaoSimuladasSistema.str_inicio,
                         "",
                         True)

        self._dados: pd.DataFrame = pd.DataFrame()

    def __eq__(self, o: object):
        if not isinstance(o, BlocoGeracaoUsinasNaoSimuladasSistema):
            return False
        bloco: BlocoGeracaoUsinasNaoSimuladasSistema = o
        return self._dados.equals(bloco._dados)

    # Override
    def le(self, arq: IO):

        def converte_tabela_em_df() -> pd.DataFrame:
            df = pd.DataFrame(tabela)
            df.columns = MESES_DF
            df["Ano"] = anos
            df["Subsistema"] = subsistema
            df = df[["Ano", "Subsistema"] + MESES_DF]
            return df

        # Variáveis auxiliares
        reg_subsis = RegistroIn(3)
        reg_ano = RegistroIn(4)
        reg_ger = RegistroFn(7)
        # Pula duas linhas, com cabeçalhos
        arq.readline()
        arq.readline()
        i = 0
        anos = []
        subsistema = []
        subsistema_atual = 0
        tabela = np.zeros((MAX_ANOS_ESTUDO * len(SUBMERCADOS),
                          len(MESES)))
        while True:
            # Verifica se o arquivo acabou
            linha: str = arq.readline()
            if BlocoGeracaoUsinasNaoSimuladasSistema.str_fim == linha.strip():
                tabela = tabela[:i, :]
                self._dados = converte_tabela_em_df()
                break
            # Senão, lê mais uma linha
            if len(linha.strip()) < 5:
                subsistema_atual = reg_subsis.le_registro(linha, 1)
            else:
                # Ano
                anos.append(reg_ano.le_registro(linha, 0))
                # Subsistema
                subsistema.append(subsistema_atual)
                # Limites
                tabela[i, :] = reg_ger.le_linha_tabela(linha,
                                                       7,
                                                       1,
                                                       len(MESES))
                i += 1

    # Override
    def escreve(self, arq: IO):

        def escreve_geracoes():
            lin_tab = self._dados.shape[0]
            subsistema_anterior = 0
            for i in range(lin_tab):
                linha = ""
                # Subsistema
                subsistema = self._dados.iloc[i, 1]
                if subsistema != subsistema_anterior:
                    arq.write(" " + str(subsistema).rjust(3) + "\n")
                    subsistema_anterior = subsistema
                # Mercados de cada mês
                linha = f"{str(self._dados.iloc[i, 0]).ljust(4)}   "
                for j in range(len(MESES)):
                    v = self._dados.iloc[i, j + 2]
                    linha += " {:6.0f}".format(v).rjust(7)
                arq.write(linha + "\n")

        # Escreve cabeçalhos
        arq.write(f"{BlocoGeracaoUsinasNaoSimuladasSistema.str_inicio}\n")
        arq.write(" XXX  XBL  XXXXXXXXXXXXXXXXXXXX  XTE\n")
        cab = ("       XXXJAN. XXXFEV. XXXMAR. XXXABR. XXXMAI. XXXJUN." +
               " XXXJUL. XXXAGO. XXXSET. XXXOUT. XXXNOV. XXXDEZ.\n")
        arq.write(cab)
        escreve_geracoes()
        # Escreve a linha de terminação
        arq.write(f" {BlocoGeracaoUsinasNaoSimuladasSistema.str_fim}\n")


class LeituraSistema(Leitura):
    """
    Realiza a leitura do arquivo `sistema.dat`
    existente em um diretório de entradas do NEWAVE.

    Esta classe contém o conjunto de utilidades para ler
    e interpretar os campos de um arquivo `sistema.dat`, construindo
    um objeto `Sistema` cujas informações são as mesmas do sistema.dat.

    Este objeto existe para retirar do modelo de dados a complexidade
    de iterar pelas linhas do arquivo, recortar colunas, converter
    tipos de dados, dentre outras tarefas necessárias para a leitura.
    """

    def __init__(self,
                 diretorio: str):
        super().__init__(diretorio)

    # Override
    def _cria_blocos_leitura(self) -> List[Bloco]:
        """
        Cria a lista de blocos a serem lidos no arquivo sistema.dat.
        """
        return [BlocoCustoDeficitSistema(),
                BlocoIntercambioSistema(),
                BlocoMercadoEnergiaSistema(),
                BlocoGeracaoUsinasNaoSimuladasSistema()]
